/* Shared library add-on to iptables to add SETPRIO target support. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include "../../../apps/fw_modules_priv/ipt_SETSPRIO.h"

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"SETPRIO target v%s options:\n"
"  --set-prio value                   Set nfmark value\n"
"  --[high|low|overwrite]         Comparison mode\n"
"             high: keep the higher prio\n"
"             low : keep the lower prio\n"
"             overwrite: use the user-assigned prio\n"
"  --by-[sport|dport|smac|dmac]\n"
"            sport: source port\n"
"            dport: destination port\n"
"            smac: source MAC address\n"
"            dmac: destination MAC address\n"
"\n",
IPTABLES_VERSION);
}

static struct option opts[] = {
	//{name, has_arg, flag, parse_val}
	// 1 means it must come with a sub-param
	{"set-prio", 1, 0, 'p' }, 
	{"high",0,0,'h'},
	{"low",0,0,'l'},
	{"overwrite",0,0,'o'},
	{"by-dport",0,0,'d'},
	{"by-sport",0,0,'s'},
	{"by-dmac",0,0,'m'},
	{"by-smac",0,0,'z'},
	{"by-ip",0,0,'i'},
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_setsprio_info *setsprioinfo = (struct ipt_setsprio_info *)t->data;
	setsprioinfo->sprio=SPRIO_NULL; //default PRIO ---------------------------------------------->to be fixed
	//*nfcache |= NFC_ALTERED;
	*nfcache |= NFC_UNKNOWN;
	printf("SETSPRIO initialized.\n");
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
 	//this function will be run several times depends on how many params were passed

	struct ipt_setsprio_info *setsprioinfo = (struct ipt_setsprio_info *)(*target)->data;

	switch (c) {
		case 'p':
			if (string_to_number(optarg, 0, 256, (unsigned int *)&setsprioinfo->sprio))
				exit_error(PARAMETER_PROBLEM, "Bad --set-prio value `%s'", optarg);
			*flags |= FLAG_DONE_PRIO;
			break;

		case 'h': //high
			setsprioinfo->mode=SPRIO_HIGH;
			*flags |= FLAG_DONE_MODE;
			break;

		case 'l': //Low
			setsprioinfo->mode=SPRIO_LOW;
			*flags |= FLAG_DONE_MODE;
			break;
		case 'o': //overwrite
			setsprioinfo->mode=SPRIO_OVERWRITE;
			*flags |= FLAG_DONE_MODE;
			break;
		case 'd':
			setsprioinfo->table_type=SPRIO_TABLE_DPORT;
			*flags |= FLAG_DONE_TYPE;
			break;
		case 's':
			setsprioinfo->table_type=SPRIO_TABLE_SPORT;
			*flags |= FLAG_DONE_TYPE;
			break;
		case 'm':
			setsprioinfo->table_type=SPRIO_TABLE_DMAC;
			*flags |= FLAG_DONE_TYPE;
			break;
		case 'z':
			setsprioinfo->table_type=SPRIO_TABLE_SMAC;
			*flags |= FLAG_DONE_TYPE;
			break;

		case 'i':
			setsprioinfo->table_type=SPRIO_TABLE_IP;
			*flags |= FLAG_DONE_TYPE;
			break;
/*
		default:
			*flags=0;
			return 0;
			break;
*/
	}

	//this section makes sure the non-set information becomes SPRIO_NULL
	//this section is some what inefficient since it will be executed several times
	if (!(*flags & FLAG_DONE_PRIO))
		setsprioinfo->sprio=SPRIO_NULL;//set null first
	if (!(*flags & FLAG_DONE_TYPE))
		setsprioinfo->table_type=SPRIO_NULL; //default
	if (!(*flags & FLAG_DONE_MODE))
		setsprioinfo->mode = SPRIO_NULL;//default

	return 1; //return 1 is to tell iptables that "yes, the param is for me".
}

static void
final_check(unsigned int flags){
	if (flags==0)  
		exit_error(PARAMETER_PROBLEM, "SETSPRIO:flag either --set-prio  or --by-[dport|sport|mac]   is required.");

	if ((flags & FLAG_DONE_PRIO) && (flags & FLAG_DONE_TYPE))
		exit_error(PARAMETER_PROBLEM, "SETSPRIO: --set-prio  and --by-[dport|sport|mac]  MUST NOT co-exist!.");		
}

static void print_setsprio(uint prio, uint ttype, uint mode,int numeric){
	printf("sprio:%d    type:%d   mode:%d",prio,ttype,mode);
}

/* Prints out the targinfo. */
static void print(const struct ipt_ip *ip, const struct ipt_entry_target *target, int numeric){
  //this fxn is for   iptables -L -t mangle
	const struct ipt_setsprio_info *setsprioinfo = 	(const struct ipt_setsprio_info *)target->data;
	print_setsprio(setsprioinfo->sprio, setsprioinfo->table_type,setsprioinfo->mode, numeric);
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_setsprio_info *setsprioinfo = (const struct ipt_setsprio_info *)target->data;

	printf("--set-prio 0x%ulx ", setsprioinfo->sprio);
}

static
struct iptables_target setsprio
= { NULL,
    "SETSPRIO",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_setsprio_info)),
    IPT_ALIGN(sizeof(struct ipt_setsprio_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_target(&setsprio);
}

